# Platformy Technologiczne

## Obsługa wątków

Michał Wójcik

---

### Proces

**Proces** - pojedyncza instancja wykonywanego programu. Każdy proces posiada własne:

* unikatowy numer identyfikujący (PID - process identifier),
* środowisko wykonania,
* skończony zbiór zasobów,
* przestrzeń w pamięci.

**Zasoby** przydzielone procesowi:

* czas procesora,
* pamięć,
* dostęp do urządzeń I/O,
* pliki.

Zarządzaniem procesami zajmuje się **jądro** systemu operacyjnego.

---

### Wykonanie procesu

Sposoby **wykonywania** procesów:

* sekwencyjnie - poszczególne akcje procesu są wykonywane jedna po drugiej, kolejna akcja rozpoczyna
się po zakończeniu poprzedniej,
* współbieżnie - kolejna akcja może rozpocząć się przed zakończeniem poprzedniej, akcje mogą być
wykonywane w tym samym czasie lub z przeplotem (procesor przełącza się pomiędzy różnymi akcjami),
* równolegle - akcje są wykonywane w tym samym czasie na kilku procesorach (lub rdzeniach).

---

### Wątki

**Wątki** - fragmenty programu wykonywane współbieżnie (niekoniecznie równolegle):

* istnieją wewnątrz procesu,
* współdzielą miedzy sobą zasoby procesu z wyjątkiem czasu procesora.

Przykłady **wykorzystania wątków**:

* wykonywanie tych samych obliczeń na różnych i niezależnych elementach dużego zbioru,
* wykonywanie skomplikowanych akcji w tle bez blokowania interfejsu użytkownika.

---

### Sekcja krytyczna

**Sekcja krytyczna** - fragment kodu, w którym korzysta się z zasobu nieobsługującego dostępu przez
wiele wątków naraz.

Przykłady **sekcji krytycznej**:

* zmiana wartości globalnej zmiennej przez kilka wątków,
* zapisywanie/czytanie do/z jednego zasobu (pliku) przez wiele wątków,
* manipulacja jednym zbiorem (tablica, kolekcja) przez wiele wątków.

---

### Tworzenie wątków

Dwa sposoby **tworzenia wątków**

* implementacja interfejsu `Runnable` (zalecana):
  * pozwala na dziedziczenie po innych klasach,
  * skupia się na zdefiniowaniu pracy do wykonania, a nie mechanizmie utworzenia nowego wątku;
* dziedziczenie po klasie `Thread` (niezalecane):
  * obiekty `Thread` służą do utworzenia nowego wątku w maszynie wirtualnej.

---

### Główny wątek aplikacji

Metoda `main` jest przetwarzana przez główny wątek aplikacji tworzony wraz z procesem.

```Java
public class Main {

    public static void main(String[] args) {
        System.out.println("Welcome to the main thread");
    }

}
```

---

### Uśpienie wątku

Metoda `Thread.sleep` pozwala na uśpienie wątku na zadaną liczbę ms.

```Java
public class Main {

    public static void main(String[] args) {
        try {
            Thread.sleep(3000);
        } catch (InterruptedException ex) {
            
        }
        System.out.println("Welcome to the main thread");
    }

}
```

---

### Tworzenie wątków - `Runnable`

Instancję implementacji interfejsu `Runnable` można przekazać do nowego obiektu `Thread`.

```Java
public class GoodHowler implements Runnable {
    
    @Override
    public void run() {
        System.out.println("woof woof");
    }

}
```

```jshelllanguage
Thread thread = new Thread(new GoodHowler());
thread.start();
System.out.println("Right after starting thread!");
```

---

### Tworzenie wątków - `Thread`

Klasa `Thread` powinna służyć obsłudze wątku, nie definiowaniu co powinno być zrobione.

```Java
public class BadHowler extends Thread {
    
    @Override
    public void run() {
        System.out.println("woof woof");
    }

}
```

```jshelllanguage
Thread thread = new BadHowler();
thread.start();
```

---

### Sekcja krytyczna - metoda

Synchronizowane metody nie mogą być wywołane przez więcej niż jeden wątek z poziomu tego samego
obiektu w tym samym czasie.

Wszystkie synchronizowane metody są traktowane jako jedna sekcja krytyczna.

```Java
public class Resource {
    
    public synchronized void m1() {
    }
    
    public synchronized void m2() {
    }

}
```

---

### Sekcja krytyczna - metoda statyczna

Synchronizowane metody statyczne nie mogą być wywołane przez więcej niż jeden wątek z poziomu tej
samej klasy w tym samym czasie.

Wszystkie synchronizowane metody są traktowane jako jedna sekcja krytyczna.

```Java
public class Resource {
    
    public synchronized static void m1() {
    }
    
    public synchronized static void m2() {
    }

}
```

---

### Sekcja krytyczna - fragment

Synchronizowane fragmenty kodu nie mogą być przetwarzane przez więcej niż jeden wątek.
Synchronizowanie bloków wymaga podania obiektu wykorzystywanego jako rygiel (lock). Do
synchronizacji wykorzystywana jest instancja obiektu w pamięci.

Synchronizacja całych metod jest tym samym co synchronizacja bloku całego ciała metody z podaniem
obiektu `this`.

```Java
public class Resource {
    
    public void m1() {
        synchronized (this) {
            
        }
    }

}
```

---

### Sekcja krytyczna - obiekt synchronizacji

```Java
public class LocalSynchronizedResource {
    
    private final Object lock = new Object();
    
    public void m1() {
        synchronized (lock) {//Resource objects are independent.
        }
    }

}
```

```Java
public class GlobalSynchronizedResource {
    
    private static final Object lock = new Object();
    
    public void m1() {
        synchronized (lock) {//Resource objects synchronized globally.
        }
    }

}
```

---

### Uśpienie i wybudzenie wątku - przykład

```Java
public class Sleeper implements Runnable {
    
    @Override
    public void run() {
        try {
            Thread.sleep(1000 * 60 * 60);
        } catch (InterruptedException ex) {
        }
    }

}
```

```jshelllanguage
Sleeper sleeper = new Sleeper();
Thread thread = new Thread(new Sleeper());
thread.start();

thread.interrupt();
```

---

### Przerwanie wątku - przykład

```Java
public class Sleeper implements Runnable {
    
    @Override
    public void run() {
        while (!Thread.isInterrupted()) {
            try {
                Thread.sleep(1000 * 60);
            } catch (InterruptedException ex) {
                break;
            }
        }
    }

}
```

---

### Uśpienie i wybudzenie wątku

Metoda `Thread.sleep(long ms)`:

* pozwala na uśpienie wątku na określony czas,
* brak pewności ile dokładnie wątek będzie uśpiony,
* możliwe wybudzenie za pomocą metody `Thread#interrupt()`,
* wybudzenie powoduje rzucenie wyjątku `InterruptedException`,
* możliwość sprawdzenia próby wybudzenia przez metodę `Thread.isInterrupted()`,
* pozostaje w sekcji krytycznej.

Uwaga: zarówno złapanie wyjątku `InterruptedException` jak i sprawdzenie stanu metodą
`isInterrupted()` konsumuje zdarzenie, powtórne wywołanie metody `isInterrupted()` zwróci wartość
`false`.

---

### Oczekiwanie na zakończenie wątku

```Java
public class Worker implements Runnable {
    
    int result = 0;
    
    public Worker(int value) {
        result = value;
    }
    
    @Override
    public void run() {
        result *= 2;//Epic computations.
    }
    
    public int getResult() {
        return result;
    }
}
```

Metoda `Thread#join` pozwala na oczekiwania na zakończenie danego wątku.

```jshelllanguage
Worker worker = new Worker();
Thread thread = new Thread(worker);
thread.start();
//...
thread.join();
System.out.println(worker.getResult());
```

---

### Mechanizm wait-notify - zasób

```Java
public class Resource {
    
    private String value = null;
    
    public synchronized String take() throws InterruptedException {
        while (value == null) {
            wait();//Wait, maybe someone will put sth here.
        }
        String ret = value;
        value = null;
        return ret;
    }

    public synchronized void put(String value) {
        this.value = value;
        notifyAll();
    }
   
}
```

---

### Mechanizm wait-notify - producent

```Java
public class Producer implements Runnable {
    
    private Resource resource;
    
    public Producer(Resource resource) {
        this.resource = resource;
    }
    
    @Override
    public void run() {
        resource.put("woof");
    }
}
```

---

### Mechanizm wait-notify - konsument

```Java
public class Consumer implements Runnable {
    
    private Resource resource;
    
    public Consumer(Resource resource) {
        this.resource = resource;
    }
    
    @Override
    public void run() {
        try {
            System.out.println(resource.take());
        } catch (InterruptedException ex) {
            
        }
    }
}
```

---

### Mechanizm wait-notify

Metoda `Object#wait`:

* pozwala czekać na jakiś zasób,
* nie blokuje sekcji krytycznej,
* musi być wywołana na tym samym obiekcie, na którym kod jest synchronizowany.

Metody `Object#notify` i `Object#notifyAll`:

* mogą zostać użyte do powiadomienia jednego lub kilku wątków o dostępności zasobu,
* wybudzają wątki uśpione metodą `wait`,
* muszą zostać wywołane na tym samym obiekcie, na którym kod jest synchronizowany.

---

### Synchronizowane kolekcje

Synchronizowane kolekcje (dekoratory na stworzone wcześniej instancje) można uzyskać za pomocą
metod wytwórczych z klasy `Collections`:

* `synchronizedCollection`,
* `synchronizedSet`,
* `synchronizedSortedSet`,
* `synchronizedNavigableSet`,
* `synchronizedList`,
* `synchronizedMap`,
* `synchronizedSortedMap`,
* `synchronizedNavigableMap`,

```jshelllanguage
List<String> list = new ArrayList<>();
List<String> synchroList = Collections.synchronizedList(list);
```

---

### Pomocne źródła

Uzupełnienie informacji:

* <https://www.baeldung.com/cs/process-vs-thread>
* <https://www.baeldung.com/java-runnable-vs-extending-thread>
* <https://www.baeldung.com/java-wait-and-sleep>
* <https://www.baeldung.com/java-synchronized>
* <https://www.baeldung.com/java-interrupted-exception>
* <https://www.baeldung.com/java-thread-stop>
* <https://www.baeldung.com/java-producer-consumer-problem>
