# Platformy Technologiczne

## Obsługa wejścia / wyjścia

Michał Wójcik

---

### Strumienie

**Strumienie** - ciągi danych z/do których dane mogą być pobierane/dodawane:

* **Byte streams** - strumienie bajtowe, czytanie/pisanie bajtami,
* **Character streams** - strumienie znakowe, czytanie/pisanie znakami.

Istnieje wiele rodzai strumieni:

* do/z pliku,
* do/z gniazdka sieciowego,
* do/z pamięci.

---

### Bazowe strumienie

Bazowe klasy abstrakcyjne:

* **Byte streams**:
  * `InputStream`,
  * `OutputStream`;
* **Character streams**:
  * `Reader`,
  * `Writer`.

Nie definiują źródła/celu danych a jedynie ich format (znaki lub bajty).

---

### `InputStream` - metody

Metody klasy `InputStream`:

* `int read()` - zwraca jeden bajt,
* `int read(byte[] buffer)` - wczytuje bajty do tablicy i zwraca ich liczbę,
* `int read(byte[] buffer, int offset, int length)` - dodatkowe ograniczenia,
* `int available()` - estymacja liczby bajtów pozostałych do odczytania,
* `void close()` - zamyka strumień.

---

### `OutputStream` - metody

Metody klasy `OutputStream`:

* `void write(int b)` - zapisuje bajt,
* `void write(byte[] buffer)` - zapisuje bajty z tablicy,
* `void write(byte[] buffer, int offset, int length)` - dodatkowe ograniczenia,
* `void close()` - zamyka strumień.

---

### `Reader` - metody

Metody klasy `Reader`:

* `int read()` - zwraca numer znaku,
* `int read(char[] buffer)` - wczytuje znaki do tablicy i zwraca ich liczbę,
* `int read(char[] buffer, int offset, int length)` - dodatkowe ograniczenia,
* `boolean ready()` - czy zostało coś do przeczytania,
* `void close()` - zamyka strumień.

---

### `Writer` - metody

Metody klasy `Writer`:

* `void write(int c)` - zapisuje znak o podanym numerze,
* `void write(String text)` - zapisuje podany ciąg znaków,
* `void write(String text, int offset, int length)` - dodatkowe ograniczenia,
* `void write(char[] buffer)` - zapisuje znaki z tablicy,
* `void write(char[] buffer, int offset, int length)` - dodatkowe ograniczenia,
* `void close()` - zamyka strumień.

---

### Strumienie do i z plików

Konkretne klasy dla obsługi plików:

* **Byte streams**:
  * `FileInputStream`,
  * `FileOutputStream`;
* **Character streams**:
  * `FileReader`,
  * `FileWriter`.

---

### Zapis/odczyt do/z pliku tekstowego

```jshelllanguage
try {
    Writer writer = new FileWriter("/tmp/output.txt");
    writer.write("Hello World");    
    writer.close();
} catch (IOException e) {

}
```

```jshelllanguage
try {
    Reader reader = new FileReader("/tmp/output.txt");
    
    char[] buffer = new char[11];
    reader.read(buffer);
    System.out.println(buffer);
    
    reader.close();
} catch (IOException ex) {

}
```

---

### Zapis/odczyt do/z pliku binarnego

```jshelllanguage
try {
    OutputStream os = new FileOutputStream("/tmp/output");
    os.write("Hello World".getBytes());
    os.close();
} catch (IOException ex) {

}
```

```jshelllanguage
try {
    InputStream is = new FileInputStream("/tmp/output");
    
    byte[] buffer = new byte[11];
    is.read(buffer);
    System.out.println(new String(buffer));
    
    is.close();
} catch (IOException ex) {

}
```

Pliki tekstowe to pliki binarne, które można łatwo zinterpretować.

---

### Zapis do pliku binarnego

```jshelllanguage
try {
    OutputStream os = new FileOutputStream("/tmp/output");
    byte[] buffer = {2, 32, 124, 42, 12, 32, 42, 23, 42, 12, 4};
    os.write(buffer);
    os.close();
} catch (IOException ex) {

}
```

Danych, które nie reprezentują tekstu, nie można łatwo odczytać. Można skorzystać z narzędzi:

* Okteta <https://www.kde.org/applications/utilities/okteta/>,
* HxD <https://mh-nexus.de/en/hxd/>,
* HexEd.it <https://hexed.it/>.

---

### Zamykanie strumieni - podstawowe

```jshelllanguage
OutputStream os;
try {
    os = new FileOutputStream("/tmp/output");
    os.write(12);
} catch (IOException ex) {
    
} finally {
    if (os != null) {
        try {
            os.close();
        } catch (IOException ex) {
            
        }
    }
}
```

---

### Zamykanie strumieni - uproszczone

Składnia `try-with-resource` pozwala na zdefiniowanie w ramach bloku `try` zmiennej, której klasa
implementuje interfejs `AutoCloseable`. Powoduje to wygenerowanie poprawnego mechanizmy zamknięcia
stworzonego obiektu.

```jshelllanguage
try (OutputStream os = new FileOutputStream("/tmp/output")) {
    os.write(12);
} catch (IOException ex) {
    
}
```

---

### Strumienie buforowane

**Strumienie buforowane** - usprawnienie procesu odczytu/zapisu danych:

* zmniejszenie odwołań do natywnego API,
* zrealizowane w formie dekoratorów,
* nie definiują źródła/celu danych a jedynie dodają mechanizm buforowania,
* dane czytane/zapisywane większymi partiami,
* udostępniają oryginalne metody strumieni (zoptymalizowane),
* udostępniają dodatkowe metody.

---

### Strumienie buforowane - rodzaje

**Strumienie buforowane**:

* **Byte streams**:
  * `BufferedInputStream`,
  * `BufferedOutputStream`;
* **Character streams**:
  * `BufferedReader`,
  * `BufferedWriter`.

---

### Strumienie buforowane - przykład

Strumień buforowany należy nałożyć na konkretny strumień.

```jshelllanguage
try (OutputStream os = new FileOutputStream("/tmp/test");
     BufferedOutputStream bos = new BufferedOutputStream(os)) {
    bos.write(12);
} catch (IOException ex) {

}
```

---

### Metody strumieni buforowanych

Metody klasy `BufferedOutputStream`:

* `void flush()` - opróżnienie bufora.

Metody klasy `BufferedWriter`:

* `void flush()` - opróżnienie bufora,
* `void newLine()` - wstawienie znaku nowej linii.

Metody klasy `BufferedReader`:

* `String readLine()` - wczytanie całej linii tekstu.

---

### Serializacja

**Serializacja** - konwersja obiektu do strumienia bajtów.

Strumienie pozwalające na serializację:

* `ObjectInputStream`,
* `ObjectOutputStream`,
* zrealizowane w formie dekoratorów,
* nie definiują źródła ani celu danych a jedynie możliwość serializacji/deserializacji.

---

### Serializacja - konfiguracja

Konfiguracja procesu serializacji:

* serializacja pozwala na konwersję całych obiektów na strumienie bajtów,
* obiekt musi implementować interfejs `Serializable`,
* pola z modyfikatorem `transient` nie podlegają serializacji,
* klasy powinny deklarować wersję klasy za pomocą pola `static final long serialVersionUID`.

---

### Serializacja i deserializacja

```jshelllanguage
try (OutputStream os = new FileOutputStream("/tmp/object");
        ObjectOutputStream oos = new ObjectOutputStream()) {
    oos.writeDouble(3.3);
    oos.writeObject("the blue wolf");
} catch (IOException ex) {

}
```

```jshelllanguage
try (InputStream is = new FileInputStream("/tmp/object");
        ObjectInputStream ois = new ObjectInputStream(is)) {
    System.out.println(ois.readDouble());
    System.out.println(ois.readObject());
} catch (IOException | ClassNotFoundException ex) {

}
```

---

### Serializowalny obiekt

```Java
public class Box implements Serializable {
    
    static final long serialVersionUID = 13;
    
    private String label;
    
    private double capacity;
    
    private transient String placement;
    
    //getter and setters

}
```

---

### Serializacja obiektu

```jshelllanguage
try (OutputStream os = new FileOutputStream("/tmp/object");
        ObjectOutputStream oos = new ObjectOutputStream(os)) {
    Box box = new Box();
    box.setLabel("my box");
    box.setCapacity(22);
    box.setPlacement("top");
    oos.writeObject(box);
} catch (IOException ex) {

}
```

---

### Deserializacja obiektu

```jshelllanguage
try (InputStream is = new FileInputStream("/tmp/object");
        ObjectInputStream ois = new ObjectInputStream(is)) {
    Box box = (Box) ois.readObject();
    System.out.println(box.getLabel()); // my box
    System.out.println(box.getCapacity()); // 22
    System.out.println(box.getPlacement()); // null
} catch (IOException | ClassNotFoundException ex) {
    
}
```

---

### Pomocne źródła

Uzupełnienie informacji:

* <https://www.baeldung.com/java-outputstream>
* <https://www.baeldung.com/java-filewriter>
* <https://www.baeldung.com/java-filereader>
* <https://www.baeldung.com/java-buffered-reader>
* <https://www.baeldung.com/java-serialization>
* <https://www.baeldung.com/java-serial-version-uid>
