# Platformy Technologiczne

## Java Persistence API

Michał Wójcik

---

### Java Persistence API - specyfikacja

**Java Persistence API** (JPA):

* specyfikacja dla bibliotek mapowania obiektowo-relacyjnego,
* wykorzystywana w wielu frameworkach i środowiskach:
  * Java SE, Java EE, Spring Framework, Play Framework;
* nie wymaga budowania skomplikowanych obiektów DAO (ang. Data Access Object),
* obsługuje transakcje ACID (Atomicity, Consistency, Isolation, Durability),
* niezależna od dostawcy bazy danych:
  * sterowniki JDBC dla wszystkich popularnych baz,
  * pozwala na automatyczne generowanie kodu SQL.

---

### JPA - implementacje

**JPA** jest jedynie **standardem**, istnieje kilka implementacji:

* Hibernate, od Red Hat (serwer WildFly),
* Toplink od Oracle,
* OpenJPA od Apache Software Fundation,
* EclipseLink - implementacja referencyjna od Eclipse Fundation (Eclipse IDE).

---

### JPA - wersje

Wersje JPA:

* Java Persistence Service:
  * ostatnia wersja: 2.2,
  * nierozwijana,
  * pakiet `javax.persistnece`;
* Jakarta Persistence Service:
  * aktualna wersja: 3.1,
  * rozwijana,
  * pakiet `jakarta.persistence`.

---

### Zależności Java Persistence API

Zależność do interfejsów i adnotacji zdefiniowanych w specyfikacji JPA:

```xml
<dependency>
    <groupId>javax.persistence</groupId>
    <artifactId>javax.persistence-api</artifactId>
    <version>2.2</version>
</dependency>
```

Zależność do wybranej implementacji JPA:

```xml
<dependency>
    <groupId>org.hibernate</groupId>
    <artifactId>hibernate-core</artifactId>
    <version>5.6.15.Final</version>
    <scope>runtime</scope>
</dependency>
```

Zależność do sterownika JDBC dla wybranego serwera baz danych:

```xml
<dependency>
    <groupId>com.h2database</groupId>
    <artifactId>h2</artifactId>
    <version>2.1.214</version>
    <scope>runtime</scope>
</dependency>
```

---

### Zależności Jakarta Persistence API

Zależność do interfejsów i adnotacji zdefiniowanych w specyfikacji JPA:

```xml
<dependency>
    <groupId>jakarta.persistence</groupId>
    <artifactId>jakarta.persistence-api</artifactId>
    <version>3.1.0</version>
</dependency>
```

Zależność do wybranej implementacji JPA:

```xml
<dependency>
    <groupId>org.hibernate.orm</groupId>
    <artifactId>hibernate-core</artifactId>
    <version>6.2.0.Final</version>
    <scope>runtime</scope>
</dependency>
```

Zależność do sterownika JDBC dla wybranego serwera baz danych:

```xml
<dependency>
    <groupId>com.h2database</groupId>
    <artifactId>h2</artifactId>
    <version>2.1.214</version>
    <scope>runtime</scope>
</dependency>
```

---

### Klasy encyjne

**Klasy encyjne** (entity classes):

* opisują obiekty odwzorowywane na encje w bazie danych,
* zwykłe klasy POJO (Plain Old Java Object),
* pola klasy nie mogą być publiczne,
* każdy obiekt encyjny ma jednoznacznie identyfikujący go identyfikator odwzorowany na wartość
  klucza głównego,
* złożone klucze główne mogą być reprezentowane przez oddzielne klasy, które muszą implementować
  metody `hashCode()` i `equals()`,
* oznaczone za pomocą adnotacji.

---

### Istotne adnotacje JPA - klasa

**Adnotacje** na poziomie **klasy**:

* `@Entity` – oznaczenie klasy jako encyjnej (wymagana),
* `@Table` – własności tabeli w bazie danych, np.:
  * `name` – nazwa tabeli,
  * `indexes` – dodatkowe indeksy (domyślny indeks dla klucza głównego).

---

### Istotne adnotacje JPA - pola

**Adnotacje** na poziomie **pól**:

* `@Id` – oznaczenie identyfikatora (klucza głównego),
* `@GeneratedValue` – automatyczne generowanie wartości klucza głównego,
* `@Column` – własności kolumny w bazie danych, np.:
  * `name` – nazwa kolumny,
  * `nullable` – czy pole jest wymagane,
  * `unique` – czy kolumna przechowuje unikalne wartości,
  * `updatable` – czy wartość w kolumnie można modyfikować po zapisaniu nowego wiersza w tabeli;
* `@Temporal` – wymagane dla typów `Date` i `Calendar`, umożliwia wykorzystanie bazodanowych typów
  do przechowywania
  daty/czasu, jeśli baza je oferuje (niepotrzebne dla `java.time`),
* `@Transient` – oznaczenie pól klasy, które mają zostać pominięte przy mapowaniu obiektowo-relacyjnym.

---

### Przykład klasy encyjnej

```java
@NoArgsConstructor
@EqualsAndHashCode
@ToString
@Entity
public class Book {

    @Getter
    @Id
    @GeneratedValue
    private Integer id;

    @Getter
    @Setter
    private String title;

    @Getter
    @Setter
    @Column(unique = true)
    private String isbn;

}
```

---

### Generowanie wartości klucza głównego

Atrybut `strategy` adnotacji `@GeneratedValue` określa sposób generowania wartości kluczy głównych:

* `GenerationType.IDENTITY`:
  * na podstawie wartości w kolumnie klucza głównego,
  * MySQL: `AUTO_INCREMENT`,
  * PostgreSQL: `SERIAL`,
  * MS SQL: `IDENTITY(1,1)`;
* `GenerationType.SEQUENCE`:
  * na podstawie sekwencji (lub tabeli jeśli sekwencje nie są wspierane):
  * np. Oracle Database: `CREATE SEQUENCE invoice_seq START WITH 1;`
* `GenerationType.TABLE`:
  * wartości generowane na podstawie pomocniczej tabeli;
* `GenerationType.AUTO`:
  * implementacja JPA wybiera strategię w zależności od bazy danych.

---

### Związki między encjami

**Związki** pomiędzy encjami mogą zostać odzwierciedlone na poziomie obiektów encyjnych JPA:

* jednokierunkowe – jedna z klas encyjnych zawiera odwołanie do drugiej,
* dwukierunkowe – obie klasy encyjne posiadają wzajemne odwołania,

**Adnotacje** definiujące **związki**:

* `@OneToOne`,
* `@OneToMany`,
* `@ManyToOne`
* `@ManyToMany`.

Wybrane własności adnotacji:

* `mappedBy` - oznaczenie pola będącego właścicielem związku,
* `cascade` - operacje kaskadowe na elementach zależnych od siebie.

---

### Przykład związku jednokierunkowego

```java
@NoArgsConstructor
@EqualsAndHashCode
@ToString
@Entity
public class Employee {

    @Getter
    @Id
    @GeneratedValue
    private Integer id;

    @Getter
    @Setter
    private String name;


    @Getter
    @Setter
    private String surname;
    
    @Getter
    @Setter
    @ManyToOne
    private Department department;

}
```

```java
@NoArgsConstructor
@EqualsAndHashCode
@ToString
@Entity
public class Department {

    @Getter
    @Id
    @GeneratedValue
    private Integer id;

    @Getter
    @Setter
    private String name;

}
```

---

### Związki dwukierunkowe

**Związki dwukierunkowe**:

* odwzorowanie związku pomiędzy encjami w tabeli zwykle jest jednokierunkowe:
  * związki modelowane przy użyciu kluczy obcych,
  * brak informacji o drugiej stronie związku na poziomie wartości kolumn w tabeli
* w obiektowym języku programowania dwukierunkowe związki są częstym rozwiązaniem,
* klasy encyjne JPA pozwalają na dwukierunkowe odwzorowanie związku pomiędzy encjami,
* obecność atrybutu `mappedBy` wskazuje, że pole reprezentuje drugą stronę związku modelowanego
  przez inne pole,
* strona bez `mappedBy` jest stroną właścicielską – kontroluje związek między encjami.

---

### Przykład związku dwukierunkowego

```java
@NoArgsConstructor
@EqualsAndHashCode
@ToString
@Entity
public class Employee {

    @Getter
    @Id
    @GeneratedValue
    private Integer id;

    @Getter
    @Setter
    private String name;


    @Getter
    @Setter
    private String surname;
    
    @Getter
    @Setter
    @ManyToOne
    private Department department;

}
```

```java
@NoArgsConstructor
@EqualsAndHashCode
@ToString
@Entity
public class Department {

    @Getter
    @Id
    @GeneratedValue
    private Integer id;

    @Getter
    @Setter
    private String name;

    @Getter
    @OneToMany(mappedBy = "department")
    private List<Employee> employees;

}
```

---

### Przykład związku cyklicznego

```java
@NoArgsConstructor
@EqualsAndHashCode
@ToString
@Entity
public class Employee {

    @Getter
    @Id
    @GeneratedValue
    private Integer id;

    @Getter
    @Setter
    private String name;


    @Getter
    @Setter
    private String surname;
    
    @Getter
    @Setter
    @ManyToOne
    private Employee superior;

    @Getter
    @OneToMany(mappedBy = "superior")
    private List<Employee> subordinates;

}
```

---

### Jednostka trwałości

**Jednostka trwałości** (ang. Persistence Unit):

* grupuje klasy encyjne, których obiekty odwzorowywane są na encje w bazie danych,
* aplikacja może definiować kilka jednostek trwałości (używać kilku baz danych),
* określa sposób połączenia z bazą danych:
  * nazwa hosta, port, nazwa bazy, login, hasło, sterownik JDBC,
  * źródło danych (ang. DataSource) dostarczane przez środowisko wykonawcze (np. serwer aplikacji);
* standardowo konfigurowana w pliku `persistence.xml` (w katalogu `META-INF` projektu),
* frameworki mogą udostępniać alternatywne sposoby konfiguracji, definiować ustawienia domyślne:
  * np. Spring Framework: application.properties + domyślna jednostka trwałości obejmująca wszystkie
    klasy encyjne projektu.

---

### Konfiguracja jednostki trwałości - `javax`

Konfiguracja dla Java Persistence API:

```xml
<persistence xmlns="http://xmlns.jcp.org/xml/ns/persistence"
             xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
             xsi:schemaLocation="http://xmlns.jcp.org/xml/ns/persistence
                                 http://xmlns.jcp.org/xml/ns/persistence/persistence_2_2.xsd"
             version="2.2">

    <persistence-unit name="rpgPu" transaction-type="RESOURCE_LOCAL">
        <properties>
            <property name="javax.persistence.jdbc.driver" value="org.h2.Driver"/>
            <property name="javax.persistence.jdbc.url" value="jdbc:h2:mem:rpg"/>
            <property name="javax.persistence.jdbc.user" value="admin"/>
            <property name="javax.persistence.jdbc.password" value="admin"/>

            <property name="hibernate.hbm2ddl.auto" value="update"/>
            <property name="hibernate.dialect" value="org.hibernate.dialect.H2Dialect"/>
            <property name="hibernate.show_sql" value="true"/>
            <property name="hibernate.format_sql" value="true"/>
        </properties>
    </persistence-unit>

</persistence>
```

Konfiguracja jednostki trwałości zawiera między innymi konfigurację połączenia z bazą danych. W
przypadku bazy H2 zostanie ona uruchomiona automatycznie w pamięci.

---

### Konfiguracja jednostki trwałości - `jakarta`

Konfiguracja dla Jakarta Persistence API:

```xml
<persistence xmlns="https://jakarta.ee/xml/ns/persistence"
             xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
             xsi:schemaLocation="https://jakarta.ee/xml/ns/persistence 
                                 https://jakarta.ee/xml/ns/persistence/persistence_3_0.xsd"
             version="3.0">

    <persistence-unit name="rpgPu" transaction-type="RESOURCE_LOCAL">
        <properties>
            <property name="jakarta.persistence.jdbc.driver" value="org.h2.Driver"/>
            <property name="jakarta.persistence.jdbc.url" value="jdbc:h2:mem:rpg"/>
            <property name="jakarta.persistence.jdbc.user" value="admin"/>
            <property name="jakarta.persistence.jdbc.password" value="admin"/>
            
            <property name="hibernate.hbm2ddl.auto" value="update"/>
            <property name="hibernate.dialect" value="org.hibernate.dialect.H2Dialect"/>
            <property name="hibernate.show_sql" value="true"/>
            <property name="hibernate.format_sql" value="true"/>
        </properties>
    </persistence-unit>

</persistence>

```

Konfiguracja jednostki trwałości zawiera między innymi konfigurację połączenia z bazą danych. W
przypadku bazy H2 zostanie ona uruchomiona automatycznie w pamięci.

---

### Kontekst i tożsamość bazodanowa

**Kontekst** (ang. Persistence Context):

* zbiór zarządzanych (ang. managed) obiektów encyjnych,
* zmiany w zarządzanych obiektach encyjnych są śledzone i zapisywane do bazy danych po zatwierdzeniu
  transakcji.

**Tożsamość bazodanowa** (ang. Persistence Identity):

* tożsamość obiektu encyjnego wyrażona identyfikatorem, powiązanym z istniejącym kluczem głównym w
  bazie danych.

---

### Stany obiektu encyjnego

Obiekt encyjny może znajdować się w różnych **stanach**.

| stan obiektu | związany z Persistence Context | posiada Persistent Identity |
|--------------|--------------------------------|-----------------------------|
| new          | ✗                              | ✗                           |
| managed      | ✔                              | ✔                           |
| detached     | ✗                              | ✔                           |
| removed      | ✔                              | ✔                           |

Stan *removed* - przeznaczony do usunięcia przy zatwierdzeniu transakcji (obiekt istnieje w
bazie danych, dopóki transakcja nie zakończy się sukcesem).

---

### Cykl życia obiektu encyjnego

**Cykl życia** obiektu encyjnego przedstawia przejścia pomiędzy stanami, w których może się znajdować.

```mermaid
stateDiagram
    New --> Managed : persist
    Managed --> Detached : commit/rollback
    Detached --> Managed : merge
    Managed --> Managed : refresh
    Managed --> Removed : remove
    Removed --> Managed : persist
    Removed --> IllegalArgumentException : merge
    Detached --> IllegalArgumentException : persist/remove/refresh
    Removed --> New : commit
```

---

### Entity Manager

**Entity Manager** - udostępnia podstawowe operacje zarządzania obiektami encyjnymi:

* wywoływane operacje zmieniają stan obiektu encyjnego (*new*/*managed*/*detached*/*removed*),
* `void persist(Object o)` – utrwalenie obiektu (*new* ⇨ *managed*),
* `<T> T merge(T entity)` – *detached* ⇨ *managed*
* `void remove(Object o)` – usunięcie obiektu (*managed* ⇨ *removed*)
* `void refresh(Object o)` – aktualizuje stan obiektu encyjnego na podstawie bazy (*managed* ⇨ *managed*),
* `<T> T find(Class<T> entityClass, Object key)` – wyszukiwanie na podstawie wartości klucza głównego,
* `EntityTransaction getTransaction()` – zwraca obiekt transakcji:
  * większość frameworków umożliwia automatyczne zarządzanie transakcjami.

**Entity Transaction**:

* `begin()` – rozpoczyna transakcję,
* `commit()` – zatwierdza transakcję, zapisuje zmiany do bazy, odłącza zarządzane obiekty encyjne,
* `rollback()` - wycofuje transakcję, odłącza zarządzane obiekty encyjne.

---

### Pozyskanie instancji klasy EntityManager

Sposoby **pozyskania** obiektu **EntityManager** w Java SE:

* poprzez fabrykę,
* fabryka może być wykorzystywana przez wiele wątków,
* instancja `EntityManager` może być bezpiecznie używana tylko przez jeden wątek.

```jshelllanguage
EntityManagerFactory factory = Persistence.createEntityManagerFactory("booksPu");
```

```jshelllanguage
EntityManager em = factory.createEntityManager();
```

---

### Operacje na bazie - przykładowa klasa

Przykładowa klasa encyjna:

```java
@NoArgsConstructor
@EqualsAndHashCode
@ToString
@Entity
public class Book {

    @Getter
    @Setter
    private String title;

    @Getter
    @Setter
    @Id
    @Column(unique = true)
    private String isbn;

}
```

---

### Operacje na bazie

Uzyskanie obiektu fabryki:

```jshelllanguage
EntityManagerFactory emf = Persistence.createEntityManagerFactory("testPu");
EntityManager em;
```

Zapisanie nowego obiektu w transakcji:

```jshelllanguage
em = emf.createEntityManager();
em.getTransaction().begin();
em.persist(new Book("Ruda Sfora", "1"));
em.getTransaction().commit();
em.close();

```

Wyszukanie obiektu w bazie:

```jshelllanguage
em = emf.createEntityManager();
Book wolf = em.find(Book.class, "1");
System.out.println(wolf);
em.close();
```

Zamknięcie fabryki na koniec programu:

```jshelllanguage
emf.close();
```

---

### JPQL - Java Persistence Query Language

**Java Persistence Query Language** (JPQL):

* obiektowo zorientowany język zapytań o składni nawiązującej do języka SQL,
* opiera się na zdefiniowanym w projekcie modelu klas encyjnych,
* umożliwia odpytywanie bazy danych (`SELECT`), aktualizację (`UPDATE`) i usuwanie (`DELETE`) encji,
* zapytania automatycznie tłumaczone na język SQL:
  * z uwzględnieniem automatycznych złączeń tabel w czasie trawersowania związków między encjami.

---

### JPQL - przykładowa klasa

Przykładowa klasa encyjna:

```java
@Entity
public class Product {
    @Id
    UUID id = UUID.randomUUID();
    String name;
    String description;
    Integer price;
    Integer amount;
}
```

---

### JPQL - podstawowe zapytania

Pobranie wszystkich produktów:

```sql
SELECT p FROM Product p
```

Pobranie produktów o określonej nazwie:

```sql
SELECT p FROM Product p WHERE p.name = :name
```

---

### JPQL - przykładowe zapytania

Pobranie produktów o nazwie zgodnej z wyrażeniem regularnym – operator `LIKE`:

```sql
SELECT p FROM Product p WHERE p.name LIKE :name
```

Wyszukanie produktów, których opis zawiera frazę *VT-x* w dowolnym miejscu:

```sql
SELECT p FROM Product p WHERE p.description LIKE '% VT-x %'
```

Bez uwzględniania wielkości znaków:

```sql
SELECT p FROM Product p WHERE LOWER(p.name) LIKE LOWER('%Laptop%')
```

Pobranie produktów, których stan jest niski (poniżej 5 sztuk), posortowanych według nazw:

```sql
SELECT p FROM Product p WHERE p.amount < 5 ORDER BY p.name
```

---

### Wywołanie zapytań

Wywołanie zapytań bez parametrów:

```jshelllanguage
String queryString = "SELECT p FROM Product p";
Query query = em.createQuery(queryString, Product.class);
List<Product> products = query.getResultList();
```

Wywołanie zapytań z parametrami:

```jshelllanguage
String queryString = "SELECT p FROM Product p WHERE p.name LIKE :name";
Query query = em.createQuery(queryString, Product.class);
query.setParameter("name", "Laptop");
List<Product> products = query.getResultList();
```

Zliczanie:

```jshelllanguage
String queryString = "SELECT COUNT(p) FROM Product p WHERE p.name LIKE :name";
Query query = em.createQuery(queryString, Long.class);
query.setParameter("name", "Laptop");
Long count = query.getSingleResult();;
```

---

### Wywołanie zapytań cd

OFFSET oraz LIMIT:

```jshelllanguage
String queryString = "SELECT p FROM Product p";
Query query = em.createQuery(queryString, Product.class);
query.setFirstResult(21);
query.setMaxResults(10);
List<Product> products = query.getResultList();
```

Fluent syntax:

```jshelllanguage
em.createQuery("SELECT p FROM Product p WHERE p.name LIKE :name", Product.class)
       .setFirstResult(21)
       .setMaxResults(10)
       .setParameter("name", "Laptop")
       .getResultList();
```

Dla zapytań UPDATE oraz DELETE:,

```jshelllanguage
String queryString = "DELETE FROM Product p";
Query query = em.createQuery(queryString);
int changedRows = query.executeUpdate();
```

---

### Zapytania nazwane

Zapytania używane w wielu miejscach aplikacji można zdefiniować przy użyciu adnotacji `@NamedQuery`:

```java
@NamedQueries({
    @NamedQuery(name = "Product.findAll",
        query = "SELECT p FROM Product p"),
    @NamedQuery(name = "Product.findByName",
        query = "SELECT p FROM Product p WHERE p.name = :name")
})
@Entity
public class Product {
    //...
}
```

```jshelllanguage
Query query = em.createNamedQuery("Product.findAll");
```

---

### Pomocne źródła

Książki:

* *Java Persistence*, <https://en.wikibooks.org/wiki/Java_Persistence>.

Dokumentacja:

* RedHat, *Hibernate ORM Documentation*, <https://hibernate.org/orm/documentation/5.6/>.
* RedHat, *Hibernate ORM Documentation*, <https://hibernate.org/orm/documentation/6.2/>.

Uzupełnienie informacji:

* <https://www.baeldung.com/jpa-entities>
* <https://www.baeldung.com/jpa-composite-primary-keys>
* <https://www.baeldung.com/jpa-strategies-when-set-primary-key>
* <https://www.baeldung.com/jpa-one-to-one>
* <https://www.baeldung.com/jpa-join-column>
* <https://www.baeldung.com/jpa-joincolumn-vs-mappedby>
* <https://www.baeldung.com/jpa-queries>
