# Platformy Technologiczne

## Stream API

Michał Wójcik

---

### Wyrażenia lambda

**Wyrażenia lambda**:

* podejście funkcyjne do kodu, zamiast deklarować stanu obiektu, deklaruje się, co ma być zrobione,
* deklaracja funkcji realizującej jakieś działanie,
* w Javie jako klasy tymczasowe zawierające jedną metodę,
* deklaracja funkcji lambda za pośrednictwem interfejsu,
* lambdy wspierane przez język Java od wersji 1.8,
* do wersji języka Java 1.7 podejście funkcyjne realizowane przez klasy anonimowe.

---

### Lambda - deklaracja

**Założenia** przykładu:

* metoda `format` przyjmuje wartość tekstową do sformatowania,
* metoda `format` zwraca tekst sformatowany zgodnie z tym czego oczekuje programista,
* programista implementuje metodę `format` definiując jak ma przebiegać proces formatowania,
* instancje interfejsu można przekazać do dowolnej metody (np.: wypisującej coś na wyjście).

```java
public interface Formatter {
    String format(String value);
}
```

```jshelllanguage
public void printFormatted(String text, Formatter formatter) {
    System.out.println(formatter.format(text));
}
```

---

### Zwykła implementacja

Przy zwykłej implementacji do metody `printFormatted` przekazywana jest instancja konkretnej klasy
spełniającej interfejs `Formatter`.

```java
public class NamedFormatter implements Formatter {

    @Override
    public String format(String text) {
        return text + " formatter by named formatter";
    }

}
```

```jshelllanguage
printFormatted("some text", new NamedFormatter());
```

```jshelllanguage
Formatter formatter= new NamedFormatter();

printFormatted("some text", formatter);
```

---

### Klasa anonimowa

**Klasa anonimowa**:

* nie posiada nazwy,
* definiowana w kodzie (nie w osobnym pliku),
* ma dokładnie tylko jedną instancję,
* pozwala np.: na zaimplementowanie interfejsu do jednorazowego użytku,
* instancja tworzona np.: poprzez wywołanie konstruktora na interfejsie (lub klasie) i podanie implementacji.

```jshelllanguage
Formatter formatter = new Formatter() {

    @Override
    public String format(String text) {
        return text + " formatter by anonymous formatter";
    }

};

printFormatted("some text", formatter);
```

---

### Klasa anonimowa - zagnieżdżona

Przypisywanie do zmiennej pomocniczej nie jest potrzebne.

```jshelllanguage
printFormatted("some text", new Formatter() {

    @Override
    public String format(String text) {
        return text + " formatter by anonymous formatter";
    }

});
```

---

### Implementacja jako lambda

Zamiast z klasy anonimowej dla interfejsów funkcyjnych (jedna metoda abstrakcyjna) można skorzystać z
wyrażenia lambda.

```jshelllanguage
Formatter formatter = text -> {
    return text + " formatted by lambda with return";
};

printFormatted("some text", formatter);
```

W przypadku jednolinijkowej implementacji można skorzystać z formy skróconej.

```jshelllanguage
Formatter formatter = text -> text + " formatted by lambda without return";

printFormatted("some text", formatter);
```

---

### Implementacja jako lambda - zagnieżdżona

Przypisywanie do zmiennej pomocniczej nie jest potrzebne.

```jshelllanguage
printFormatted("some text", text -> {
    return text + " formatted by lambda with return";
});

```

W przypadku jednolinijkowej implementacji można skorzystać z formy skróconej.

```jshelllanguage
printFormatted("some text", text -> text + " formatted by lambda without return");
```

---

### Przykład wątku jakby wyrażenia lambda

```jshelllanguage
new Thread(new Runnable() {

    @Override
    public void run () {
        System.out.println("working");
    }

}).start();
```

```jshelllanguage
new Thread(() -> {
    System.out.println("working");
}).start();
```

```jshelllanguage
new Thread(() -> System.out.println("working")).start();
```

---

### Strumienie kolekcji

**Java Streams** - strumieniowanie kolekcji:

* nie mylić ze strumieniami wejściowymi/wyjściowymi,
* traktowanie kolekcji jako strumienia danych,
* wsparcie w języku Java od wersji 1.8,
* wykonywanie operacji na strumieniu danych:
  * filtrowanie,
  * odwzorowanie,
  * sortowanie,
  * itd;
* duże wykorzystanie podejścia funkcyjnego poprzez deklarowanie jak ma być wykonana dana operacja,
* wykorzystanie fluent syntax - wywoływanie metod łańcuchowo.

---

### Elementy strumienia

**Strumień** składa się z następujących elementów:

* uzyskanie strumienia,
* dowolna liczba metod modyfikujących,
* zakończenie strumienia.

---

### Przykładowy strumień

```jshelllanguage
Collection<String> list = List.of("ookami", "kitsune");

Stream<String> stream = list.stream();

stream = stream.filter(value -> "ookami".equals(value)); //Filter

stream = stream.sorted((o1, o2) -> o1.compareTo(o2)); //Comparator

Stream<Integer> intStream = stream.map(value -> value.length()); //Function

stream.forEach(value -> System.out.println(value)); //Supplier
```

```jshelllanguage
Collection<String> list = List.of("ookami", "kitsune");

Stream<String> stream = list.stream();

List<String> collected = stream.Collect(Collectors.toList());
```

---

### Fluent syntaxt

```jshelllanguage
List.of("ookami", "kitsune")
        .stream()
        .filter(value -> "ookami".equals(value))
        .sorted((o1, o2) -> o1.compareTo(o2))
        .map(value -> value.length())
        .forEach(value -> System.out.println(value));
```

```jshelllanguage
List<String> collected = List.of("ookami", "kitsune")
        .stream()
        .collect(Collectors.toList());
```

---

### Pomocne źródła

Uzupełnienie informacji:

* <https://www.baeldung.com/java-from-anonymous-class-to-lambda>
* <https://www.baeldung.com/java-lambdas-vs-anonymous-class>
* <https://www.baeldung.com/java-8-lambda-expressions-tips>
* <https://www.baeldung.com/java-8-functional-interfaces>
